import torch
from megastep import modules, core, plotting, spaces, scene, cubicasa
from rebar import arrdict, dotdict
import matplotlib.pyplot as plt
import numpy as np
import matplotlib as mpl

CLEARANCE = 1.

@dotdict.mapping
def expand(x):
    B, A = x.shape[:2]
    return x.reshape(B*A, 1, *x.shape[2:])

@dotdict.mapping
def collapse(x, n_agents):
    B = x.shape[0]
    return x.reshape(B//n_agents, n_agents, *x.shape[2:])

class Deathmatch:

    def __init__(self, n_envs, n_agents, *args, **kwargs):
        geometries = cubicasa.sample(max(n_envs//4, 1))
        scenery = scene.scenery(geometries, n_agents)
        self.core = core.Core(scenery, *args, res=4*128, fov=70, **kwargs)
        self._rgb = modules.RGB(self.core, n_agents=1, subsample=4)
        self._depth = modules.Depth(self.core, n_agents=1, subsample=4)
        self._imu = modules.IMU(self.core, n_agents=1)
        self._movement = modules.MomentumMovement(self.core, n_agents=1)
        self._spawner = modules.RandomSpawns(geometries, self.core)

        self.action_space = self._movement.space
        self.obs_space = dotdict.dotdict(
            rgb=self._rgb.space,
            d=self._depth.space,
            imu=self._imu.space,
            health=spaces.MultiVector(1, 1))

        self._bounds = arrdict.torchify(np.stack([g.masks.shape*g.res for g in geometries])).to(self.core.device)
        self._health = self.core.agent_full(np.nan)
        self._damage = self.core.agent_full(np.nan)

        self.n_envs = self.core.n_envs*self.core.n_agents
        self.device = self.core.device

    def _reset(self, reset=None):
        reset = (self._health <= 0) if reset is None else reset
        self._spawner(reset)
        self._health[reset] = 1.
        self._damage[reset] = 0.
        return reset.reshape(-1)

    def _shoot(self, opponents):
        res = opponents.size(-1)
        middle = slice(res//2-1, res//2+1)
        agents = torch.arange(self.core.n_agents, device=self.core.device)
        matchings = (opponents[:, :, None] == agents[None, None, :, None, None])[..., middle].any(-1).any(-1)
        self.matchings = matchings
        
        hits = matchings.sum(2).float()
        wounds = matchings.sum(1).float()

        self._damage[:] += .05*hits

        pos = self.core.agents.positions 
        outside = (pos < -CLEARANCE).any(-1) | (pos > (self._bounds[:, None] + CLEARANCE)).any(-1)

        # 5% damage per hit, .1% damage per timestep
        self._health[:] += -.05*(wounds + outside) - .001
        
        return hits.reshape(-1)

    def _observe(self):
        r = modules.render(self.core)
        line_idxs = modules.downsample(r.indices, self._rgb.subsample)[..., self._rgb.subsample//2]
        obj_idxs = line_idxs//len(self.core.scenery.model)
        mask = (0 <= line_idxs) & (obj_idxs < self.core.n_agents)
        opponents = obj_idxs.where(mask, torch.full_like(line_idxs, -1))
        hits = self._shoot(opponents)
        obs = arrdict.arrdict(
                rgb=self._rgb(r), 
                d=self._depth(r), 
                imu=self._imu(),
                health=self._health.unsqueeze(-1).clone())
        return obs, hits

    @torch.no_grad()
    def reset(self):
        reset = self._reset(self.core.agent_full(True))
        obs, reward = self._observe()
        return arrdict.arrdict(
            obs=expand(obs),
            reward=reward,
            reset=reset)

    @torch.no_grad()
    def step(self, decision):
        reset = self._reset()
        self._movement(collapse(decision, self.core.n_agents))
        obs, reward = self._observe()
        return arrdict.arrdict(
            obs=expand(obs),
            reward=reward,
            reset=reset)

    def state(self, e=0):
        return arrdict.arrdict(
            core=self.core.state(e),
            rgb=self._rgb.state(e),
            d=self._depth.state(e),
            health=self._health[e].clone(),
            damage=self._damage[e].clone(),
            matchings=self.matchings[e].clone(),
            bounds=self._bounds[e].clone())

    @classmethod
    def plot_state(cls, state):
        n_agents = state.core.n_agents
        show_value = 'decision' in state

        fig = plt.figure()
        gs = plt.GridSpec(n_agents, 4 if show_value else 3, fig)

        colors = [f'C{i}' for i in range(n_agents)]

        plan = core.Core.plot_state(state.core, plt.subplot(gs[:-1, :-1]))

        # Add hits
        origin, dest = state.matchings.nonzero()
        lines = state.core.agents.positions[np.stack([origin, dest], 1)]
        linecolors = np.array(colors)[origin]
        lines = mpl.collections.LineCollection(lines, color=linecolors, linewidth=1, alpha=.5)
        plan.add_collection(lines)

        # Add bounding box
        size = state.bounds[::-1] + 2*CLEARANCE
        bounds = mpl.patches.Rectangle(
            (-CLEARANCE, -CLEARANCE), *size, 
            linewidth=1, edgecolor='k', facecolor=(0., 0., 0., 0.))
        plan.add_artist(bounds)

        # Add observations
        images = {'rgb': state.rgb, 'd': state.d}
        plotting.plot_images(images, [plt.subplot(gs[i, -1]) for i in range(n_agents)])

        ax = plt.subplot(gs[-1, 0])
        ax.barh(np.arange(n_agents), state.health, color=colors)
        ax.set_ylabel('health')
        ax.set_yticks([])
        ax.invert_yaxis()
        ax.set_xlim(0, 1)

        ax = plt.subplot(gs[-1, 1])
        ax.barh(np.arange(n_agents), state.damage, color=colors)
        ax.set_ylabel('inflicted')
        ax.set_yticks([])
        ax.invert_yaxis()

        if show_value:
            ax = plt.subplot(gs[-1, 2])
            ax.barh(np.arange(n_agents), state.decision.value, color=colors)
            ax.set_ylabel('value')
            ax.set_yticks([])
            ax.invert_yaxis()

        return fig

    def display(self, e=0):
        return self.plot_state(arrdict.numpyify(self.state(e=e)))


import av
from io import BytesIO
import numpy as np
import base64
from IPython.display import display, HTML
from pathlib import Path
from multiprocessing import cpu_count
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
from .parallel import parallel
import logging
import time
import multiprocessing
from matplotlib import tight_bbox

log = logging.getLogger(__name__)

def adjust_bbox(fig):
    bbox = fig.get_tightbbox(fig.canvas.get_renderer())
    tight_bbox.adjust_bbox(fig, bbox, fig.canvas.fixed_dpi)

def array(fig):
    adjust_bbox(fig)
    fig.canvas.draw_idle()
    renderer = fig.canvas.get_renderer()
    w, h = int(renderer.width), int(renderer.height)
    # Resolution must even, else libx264 gets upset
    h2, w2 = 2*(h//2), 2*(w//2)
    return (np.frombuffer(renderer.buffer_rgba(), np.uint8)
                        .reshape((h, w, 4))
                        [:h2, :w2, :3]
                        .copy())

class Encoder:

    def __init__(self, fps):
        """This follows the [PyAV cookbook](http://docs.mikeboers.com/pyav/develop/cookbook/numpy.html#generating-video)"""
        self._fps = fps
        self._initialized = False

    def _initialize(self, arr):
        self._content = BytesIO()
        self._container = av.open(self._content, 'w', 'mp4')

        self._stream = self._container.add_stream('h264', rate=self._fps)
        self._stream.pix_fmt = 'yuv420p'
        self._stream.height = arr.shape[0]
        self._stream.width = arr.shape[1]

        pixelformats = {1: 'gray', 3: 'rgb24'}
        self._pixelformat = pixelformats[arr.shape[2]]

        self.height, self.width = arr.shape[:2]
        self.mimetype = 'mp4'

        self._initialized = True
    
    def __enter__(self):
        return self

    def __call__(self, arr):
        if isinstance(arr, plt.Figure):
            fig = arr
            arr = array(fig)
            fig.gcf()

        if not self._initialized:
            self._initialize(arr)

        # Float arrs are assumed to have a domain of [0, 1], for backward-compatability with OpenCV.
        if np.issubdtype(arr.dtype, np.floating):
            arr = (255*arr)
        if not np.issubdtype(arr.dtype, np.uint8):
            arr = arr.astype(np.uint8).clip(0, 255)

        frame = av.VideoFrame.from_ndarray(arr, format=self._pixelformat)
        self._container.mux(self._stream.encode(frame))

    def __exit__(self, type, value, traceback):
        # Flushing the stream here causes a deprecation warning in ffmpeg
        # https://ffmpeg.zeranoe.com/forum/viewtopic.php?t=3678
        # It's old and benign and possibly only apparent in homebrew-installed ffmpeg?
        if not type:
            if hasattr(self, '_container'):
                self._container.mux(self._stream.encode())
                self._container.close()
                self.value = self._content.getvalue()
        
def html_tag(video, height=None, **kwargs):
    video = video.value if isinstance(video, Encoder) else video
    style = f'style="height: {height}px"' if height else ''
    b64 = base64.b64encode(video).decode('utf-8')
    return f"""
<video controls autoplay loop {style}>
    <source type="video/mp4" src="data:video/mp4;base64,{b64}">
    Your browser does not support the video tag.
</video>"""

def notebook(video, height=640):
    return display(HTML(html_tag(video, height)))

def _init():
    # Suppress keyboard interrupt of workers, since exiting the context 
    # manager in the parent will shut them down.
    import signal
    signal.signal(signal.SIGINT, lambda h, f: None)

def _array(f, *args, **kwargs):
    result = f(*args, **kwargs)
    if isinstance(result, plt.Figure):
        arr = array(result)
        plt.close(result)
        return arr
    else:
        return result

class ParallelEncoder:

    def __init__(self, f, fps=20, N=None):
        N = multiprocessing.cpu_count()//4 if N is None else N
        self._encoder = Encoder(fps)
        self._f = f
        self._pool = parallel(_array, progress=False, N=N, initializer=_init)
        self._queuelen = N

    def __enter__(self):
        self._futures = {}
        self._submitted = 0
        self._contiguous = 0

        self._encoder.__enter__()
        self._submit = self._pool.__enter__()
        return self

    def _process_done(self):
        while True:
            if (self._contiguous in self._futures) and self._futures[self._contiguous].done():
                result = self._futures[self._contiguous].result()
                self._encoder(result)
                del self._futures[self._contiguous]
                self._contiguous += 1
            else:
                break

    def _wait(self):
        while self._futures:
            self._process_done()
            time.sleep(.01)

    def __exit__(self, t, v, tb):
        self._wait()
        self._encoder.__exit__(t, v, tb)
        self._pool.__exit__(t, v, tb)

    def __call__(self, *args, **kwargs):
        while len(self._futures) > self._queuelen:
            self._process_done()

        self._futures[self._submitted] = self._submit(self._f, *args, **kwargs)
        self._submitted += 1
        self._process_done()

    def result(self):
        self._wait()
        return self._encoder.value

    def notebook(self):
        return notebook(self.result())

    def save(self, path):
        Path(path).write_bytes(self.result())

def encode(f, *indexable, fps=20, N=0, n_frames=None, **kwargs):
    """To use this with N > 0, you need to return an array and - if it's a new figure each time - 
    close it afterwards"""
    n_frames = len(indexable[0]) if n_frames is None else n_frames
    log.info(f'Encoding begun on {n_frames} frames')
    queuesize = 2*cpu_count() 
    submitted, contiguous = 0, 0
    futures = {}
    with Encoder(fps) as encoder, parallel(f, progress=False, N=N) as p, tqdm(total=n_frames) as pbar:
        while True:
            if (submitted < n_frames) and (len(futures) < queuesize):
                futures[submitted] = p(*[iable[submitted] for iable in indexable], **kwargs)
                submitted += 1
            if (contiguous in futures) and futures[contiguous].done():
                result = futures[contiguous].result()
                if isinstance(result, plt.Figure):
                    fig = result
                    result = array(fig)
                    plt.close(fig)
                encoder(result)
                del futures[contiguous]
                contiguous += 1
                pbar.update(1)
                if (N == 0) and (contiguous % 100 == 0):
                    log.info(f'Finished {contiguous}/{n_frames} frames')
            if contiguous == n_frames:
                log.info('Encoding finished')
                return encoder